/********************************************************************************
    Copyright (C) 1999  Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

/* NOTE: do not compile this file. It is included inline into vdecoder.cc. */

#include "ulib/utility/bitstream/membitsread.hh"
#include "video12/vlc.hh"
#include "types.hh"
#include "error.hh"


#define maybeinline inline

struct DirectVLCTableEntry
{
  short bits; int nBits;
};

struct DirectVLC
{
  int    value;
  uint16 bits;
};




/************************************* MB-INCR ********************************/



static DirectVLCTableEntry vlc_mbaddrinc_tab[] =
{
  {0x08,11}, // ESCAPE-CODE

  /* table taken from MSSG-encoder */
  {0x01,1},  {0x03,3},  {0x02,3},  {0x03,4},
  {0x02,4},  {0x03,5},  {0x02,5},  {0x07,7},
  {0x06,7},  {0x0b,8},  {0x0a,8},  {0x09,8},
  {0x08,8},  {0x07,8},  {0x06,8},  {0x17,10},
  {0x16,10}, {0x15,10}, {0x14,10}, {0x13,10},
   {0x12,10}, {0x23,11}, {0x22,11}, {0x21,11},
  {0x20,11}, {0x1f,11}, {0x1e,11}, {0x1d,11},
  {0x1c,11}, {0x1b,11}, {0x1a,11}, {0x19,11},
  {0x18,11}, {0,-1}
};


#include <iostream.h>
#include <iomanip.h>

static DirectVLC vlc_mbaddr[2048 /*2^11*/];

maybeinline int GetMBAddrIncr(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(11);
  int val = vlc_mbaddr[code].value;

  bs.SkipBits(vlc_mbaddr[code].bits);

  if (val<0)
    throw Excpt_Base(ErrSev_Warning,"Invalid MPEG stream, not existing MB-addr-incr. read. (1)");

  if (val!=0) return val;

  int val1=33;
loop:
  code = bs.PeekBits(11);
  val = vlc_mbaddr[code].value;
  bs.SkipBits(vlc_mbaddr[code].bits);

  if (val<0)
    throw Excpt_Base(ErrSev_Warning,"Invalid MPEG stream, not existing MB-addr-incr. read. (2)");

  if (val!=0) return val+val1;
  else        { val1 += 33; goto loop; }
}





/************************************* MB-MODE ********************************/





static DirectVLCTableEntry vlc_mbmode_P_tab[] =
{
  {0,0}, {3,5}, {1,2}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
  {1,3}, {0,0}, {1,1}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
  {0,0}, {1,6}, {1,5}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
  {0,0}, {0,0}, {2,5}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,-1}
};

static DirectVLCTableEntry vlc_mbmode_B_tab[] =
{
  {0,0}, {3,5}, {0,0}, {0,0}, {2,3}, {0,0}, {3,3}, {0,0},
  {2,4}, {0,0}, {3,4}, {0,0}, {2,2}, {0,0}, {3,2}, {0,0},
  {0,0}, {1,6}, {0,0}, {0,0}, {0,0}, {0,0}, {2,6}, {0,0},
  {0,0}, {0,0}, {3,6}, {0,0}, {0,0}, {0,0}, {2,5}, {0,0}, {0,-1}
};


static DirectVLC vlc_mbmode_P[64 /*2^6*/];
static DirectVLC vlc_mbmode_B[64 /*2^6*/];


static int GetMBMode_I(class MemBitstreamReader& bs)
{
  int code = bs.PeekBits(2);
  if (code & 2)
    {
      bs.SkipBits(1);
      return 1;
    }
  else
    {
      if (code!=1)
        throw Excpt_Base(ErrSev_Warning,"Invalid MPEG stream, not existing MB-class (I-picture) read.");

      bs.SkipBits(2);
      return 17;
    }
}

static int GetMBMode_P(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(6);
  bs.SkipBits(vlc_mbmode_P[code].bits);

  if (vlc_mbmode_P[code].value==0)
    throw Excpt_Base(ErrSev_Warning,"Invalid MPEG stream, not existing MB-class (P-picture) read.");

  return vlc_mbmode_P[code].value;
}

static int GetMBMode_B(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(6);
  bs.SkipBits(vlc_mbmode_B[code].bits);

  if (vlc_mbmode_B[code].value==0)
    {
      cerr << "CODEVAL: $" << hex << code << " 6 bits " << endl;
      throw Excpt_Base(ErrSev_Warning,"Invalid MPEG stream, not existing MB-class (B-picture) read.");
    }

  return vlc_mbmode_B[code].value;
}

static int GetMBMode_D(class MemBitstreamReader& bs)
{
  int c = bs.GetBits(1);
  if (c != 1)
    {
      throw Excpt_Base(ErrSev_Error,"Invalid MPEG stream, invalid MB-mode in D-picture read.");
    }

  return MBMODE_INTRA;
}

static int (* GetMBMode[5])(class MemBitstreamReader&) =
{
  NULL,
  GetMBMode_I,
  GetMBMode_P,
  GetMBMode_B,
  GetMBMode_D
};



/************************************* MotionCode ********************************/

static DirectVLCTableEntry vlc_mcode_tab[]=
{
  /* table taken from MSSG-encoder */

  {0x01,1},  {0x01,2},  {0x01,3},  {0x01,4},
  {0x03,6},  {0x05,7},  {0x04,7},  {0x03,7},
  {0x0b,9},  {0x0a,9},  {0x09,9},  {0x11,10},
  {0x10,10}, {0x0f,10}, {0x0e,10}, {0x0d,10},
  {0x0c,10}, {0,-1}
};

static DirectVLC vlc_mc[1024 /*2^10*/];

maybeinline int GetMotionCode(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(10);
  bs.SkipBits(vlc_mc[code].bits);

  if (vlc_mc[code].value==0)
    return 0;

  if (bs.GetBits(1)==0)
    return  vlc_mc[code].value;
  else
    return -vlc_mc[code].value;
}


/************************************* CBP ********************************/

static DirectVLCTableEntry vlc_cbp_tab[]=
{
  /* table taken from MSSG-encoder */

  {0x01,9}, {0x0b,5}, {0x09,5}, {0x0d,6}, 
  {0x0d,4}, {0x17,7}, {0x13,7}, {0x1f,8}, 
  {0x0c,4}, {0x16,7}, {0x12,7}, {0x1e,8}, 
  {0x13,5}, {0x1b,8}, {0x17,8}, {0x13,8}, 
  {0x0b,4}, {0x15,7}, {0x11,7}, {0x1d,8}, 
  {0x11,5}, {0x19,8}, {0x15,8}, {0x11,8}, 
  {0x0f,6}, {0x0f,8}, {0x0d,8}, {0x03,9}, 
  {0x0f,5}, {0x0b,8}, {0x07,8}, {0x07,9}, 
  {0x0a,4}, {0x14,7}, {0x10,7}, {0x1c,8}, 
  {0x0e,6}, {0x0e,8}, {0x0c,8}, {0x02,9}, 
  {0x10,5}, {0x18,8}, {0x14,8}, {0x10,8}, 
  {0x0e,5}, {0x0a,8}, {0x06,8}, {0x06,9}, 
  {0x12,5}, {0x1a,8}, {0x16,8}, {0x12,8}, 
  {0x0d,5}, {0x09,8}, {0x05,8}, {0x05,9}, 
  {0x0c,5}, {0x08,8}, {0x04,8}, {0x04,9},
  {0x07,3}, {0x0a,5}, {0x08,5}, {0x0c,6}, {0,-1}
};

static DirectVLC vlc_cbp[512 /*2^9*/];


maybeinline int GetCBP(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(9);
  bs.SkipBits(vlc_cbp[code].bits);
  return vlc_cbp[code].value;
}


/************************************* Initialization ********************************/

static DirectVLCTableEntry vlc_DClum_tab[]=
{
  /* table taken from MSSG-encoder */
  {0x0004,3}, {0x0000,2}, {0x0001,2}, {0x0005,3}, {0x0006,3}, {0x000e,4},
  {0x001e,5}, {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x01ff,9}, {0,-1}
};

static DirectVLCTableEntry vlc_DCchrom_tab[]=
{
  /* table taken from MSSG-encoder */
  {0x0000,2}, {0x0001,2}, {0x0002,2}, {0x0006,3}, {0x000e,4}, {0x001e,5},
  {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x03fe,10},{0x03ff,10}, {0,-1}
};

static DirectVLC vlc_DClum  [1<< 9];
static DirectVLC vlc_DCchrom[1<<10];

maybeinline int GetDClumSize(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(9);
  bs.SkipBits(vlc_DClum[code].bits);
  return vlc_DClum[code].value;
}

maybeinline int GetDCchromSize(class MemBitstreamReader& bs)
{
  uint16 code = bs.PeekBits(10);
  bs.SkipBits(vlc_DCchrom[code].bits);
  return vlc_DCchrom[code].value;
}


/************************************* Run/Level pairs ********************************/

struct DCTtab {
  char run, level, len;
};


/* Table B-14, DCT coefficients table zero,
 * codes 0100 ... 1xxx (used for first (DC) coefficient)
 */
static DCTtab DCTtabfirst[12] =
{
  {0,2,4}, {2,1,4}, {1,1,3}, {1,1,3},
  {0,1,1}, {0,1,1}, {0,1,1}, {0,1,1},
  {0,1,1}, {0,1,1}, {0,1,1}, {0,1,1}
};

/* Table B-14, DCT coefficients table zero,
 * codes 0100 ... 1xxx (used for all other coefficients)
 */
static DCTtab DCTtabnext[12] =
{
  {0,2,4},  {2,1,4},  {1,1,3},  {1,1,3},
  {64,0,2}, {64,0,2}, {64,0,2}, {64,0,2}, /* EOB */
  {0,1,2},  {0,1,2},  {0,1,2},  {0,1,2}
};

/* Table B-14, DCT coefficients table zero,
 * codes 000001xx ... 00111xxx
 */
static DCTtab DCTtab0[60] =
{
  {65,0,6}, {65,0,6}, {65,0,6}, {65,0,6}, /* Escape */
  {2,2,7}, {2,2,7}, {9,1,7}, {9,1,7},
  {0,4,7}, {0,4,7}, {8,1,7}, {8,1,7},
  {7,1,6}, {7,1,6}, {7,1,6}, {7,1,6},
  {6,1,6}, {6,1,6}, {6,1,6}, {6,1,6},
  {1,2,6}, {1,2,6}, {1,2,6}, {1,2,6},
  {5,1,6}, {5,1,6}, {5,1,6}, {5,1,6},
  {13,1,8}, {0,6,8}, {12,1,8}, {11,1,8},
  {3,2,8}, {1,3,8}, {0,5,8}, {10,1,8},
  {0,3,5}, {0,3,5}, {0,3,5}, {0,3,5},
  {0,3,5}, {0,3,5}, {0,3,5}, {0,3,5},
  {4,1,5}, {4,1,5}, {4,1,5}, {4,1,5},
  {4,1,5}, {4,1,5}, {4,1,5}, {4,1,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5}
};

/* Table B-15, DCT coefficients table one,
 * codes 000001xx ... 11111111
*/
static DCTtab DCTtab0a[252] =
{
  {65,0,6}, {65,0,6}, {65,0,6}, {65,0,6}, /* Escape */
  {7,1,7}, {7,1,7}, {8,1,7}, {8,1,7},
  {6,1,7}, {6,1,7}, {2,2,7}, {2,2,7},
  {0,7,6}, {0,7,6}, {0,7,6}, {0,7,6},
  {0,6,6}, {0,6,6}, {0,6,6}, {0,6,6},
  {4,1,6}, {4,1,6}, {4,1,6}, {4,1,6},
  {5,1,6}, {5,1,6}, {5,1,6}, {5,1,6},
  {1,5,8}, {11,1,8}, {0,11,8}, {0,10,8},
  {13,1,8}, {12,1,8}, {3,2,8}, {1,4,8},
  {2,1,5}, {2,1,5}, {2,1,5}, {2,1,5},
  {2,1,5}, {2,1,5}, {2,1,5}, {2,1,5},
  {1,2,5}, {1,2,5}, {1,2,5}, {1,2,5},
  {1,2,5}, {1,2,5}, {1,2,5}, {1,2,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4}, /* EOB */
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,4,5}, {0,4,5}, {0,4,5}, {0,4,5},
  {0,4,5}, {0,4,5}, {0,4,5}, {0,4,5},
  {0,5,5}, {0,5,5}, {0,5,5}, {0,5,5},
  {0,5,5}, {0,5,5}, {0,5,5}, {0,5,5},
  {9,1,7}, {9,1,7}, {1,3,7}, {1,3,7},
  {10,1,7}, {10,1,7}, {0,8,7}, {0,8,7},
  {0,9,7}, {0,9,7}, {0,12,8}, {0,13,8},
  {2,3,8}, {4,2,8}, {0,14,8}, {0,15,8}
};

/* Table B-14, DCT coefficients table zero,
 * codes 0000001000 ... 0000001111
 */
static DCTtab DCTtab1[8] =
{
  {16,1,10}, {5,2,10}, {0,7,10}, {2,3,10},
  {1,4,10}, {15,1,10}, {14,1,10}, {4,2,10}
};

/* Table B-15, DCT coefficients table one,
 * codes 000000100x ... 000000111x
 */
static DCTtab DCTtab1a[8] =
{
  {5,2,9}, {5,2,9}, {14,1,9}, {14,1,9},
  {2,4,10}, {16,1,10}, {15,1,9}, {15,1,9}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 000000010000 ... 000000011111
 */
static DCTtab DCTtab2[16] =
{
  {0,11,12}, {8,2,12}, {4,3,12}, {0,10,12},
  {2,4,12}, {7,2,12}, {21,1,12}, {20,1,12},
  {0,9,12}, {19,1,12}, {18,1,12}, {1,5,12},
  {3,3,12}, {0,8,12}, {6,2,12}, {17,1,12}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 0000000010000 ... 0000000011111
 */
static DCTtab DCTtab3[16] =
{
  {10,2,13}, {9,2,13}, {5,3,13}, {3,4,13},
  {2,5,13}, {1,7,13}, {1,6,13}, {0,15,13},
  {0,14,13}, {0,13,13}, {0,12,13}, {26,1,13},
  {25,1,13}, {24,1,13}, {23,1,13}, {22,1,13}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 00000000010000 ... 00000000011111
 */
static DCTtab DCTtab4[16] =
{
  {0,31,14}, {0,30,14}, {0,29,14}, {0,28,14},
  {0,27,14}, {0,26,14}, {0,25,14}, {0,24,14},
  {0,23,14}, {0,22,14}, {0,21,14}, {0,20,14},
  {0,19,14}, {0,18,14}, {0,17,14}, {0,16,14}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 000000000010000 ... 000000000011111
 */
static DCTtab DCTtab5[16] =
{
  {0,40,15}, {0,39,15}, {0,38,15}, {0,37,15},
  {0,36,15}, {0,35,15}, {0,34,15}, {0,33,15},
  {0,32,15}, {1,14,15}, {1,13,15}, {1,12,15},
  {1,11,15}, {1,10,15}, {1,9,15}, {1,8,15}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 0000000000010000 ... 0000000000011111
 */
static DCTtab DCTtab6[16] =
{
  {1,18,16}, {1,17,16}, {1,16,16}, {1,15,16},
  {6,3,16}, {16,2,16}, {15,2,16}, {14,2,16},
  {13,2,16}, {12,2,16}, {11,2,16}, {31,1,16},
  {30,1,16}, {29,1,16}, {28,1,16}, {27,1,16}
};


maybeinline bool GetRunLen(MemBitstreamReader& bs,int& run,int& value ,bool B15,bool first,bool MPEG1)
{
  uint16 code = bs.PeekBits(16);

  //cout << "bits16: " << hex << code << dec << endl;

  DCTtab* tab;

  if (!B15 && first && code&0x8000)
    {
      // cout << "FIRST bits: " << hex << code << dec << endl;
      if (code&0x4000) value=-1;
      else             value= 1;
      run=0;

#if 0 // TODO: reenable
#ifndef NDEBUG
      if (options.Trace_DCTCoeff)
        cout << "(0," << value << ")";
#endif
#endif

      bs.SkipBits(2);
      return false;
    }

  if (code>=16384 && !B15)
    tab = &DCTtabnext[(code>>12)-4];
  else if (code>=1024)
    {
      if (B15)
        tab = &DCTtab0a[(code>>8)-4];
      else
        tab = &DCTtab0[(code>>8)-4];
    }
  else if (code>=512)
    {
      if (B15)
        tab = &DCTtab1a[(code>>6)-8];
      else
        tab = &DCTtab1[(code>>6)-8];
    }
  else if (code>=256)
    tab = &DCTtab2[(code>>4)-16];
  else if (code>=128)
    tab = &DCTtab3[(code>>3)-16];
  else if (code>=64)
    tab = &DCTtab4[(code>>2)-16];
  else if (code>=32)
    tab = &DCTtab5[(code>>1)-16];
  else if (code>=16)
    tab = &DCTtab6[code-16];
  else
    {
      MessageDisplay::Show(ErrSev_Warning,"invalid Huffman code for DCT run/level-pair");
      return false;
    }


  bs.SkipBits(tab->len);

  if (tab->run==64)  // EOB
    {
#if 0  // TODO: reenable
#ifndef NDEBUG
      if (options.Trace_DCTCoeff)
        cout << "EOB";
#endif
#endif
      return true;
    }

  if (tab->run==65) /* escape */
    {
      if (MPEG1)
        {
          run = bs.GetBits(6);
      
          value = bs.GetBits(8);
          if (value==0)
            value = bs.GetBits(8);
          else if (value==128)
            value = bs.GetBits(8) - 256;
          else if (value>128)
            value -= 256;
        }
      else
        {
          run = bs.GetBits(6);
          value = bs.GetBits(12);
          if (value==0)
            throw Excpt_Base(ErrSev_Warning,
                             "Invalid MPEG stream, not allowed MPEG-2 escape seq value 0 read.");

          if (value >= 0x800)
            value = -(2048-(value-0x800));
        }
#if 0  // TODO: reenable
#ifndef NDEBUG
      if (options.Trace_DCTCoeff)
        cout << "esc";
#endif
#endif
    }
  else
    {
      run   = tab->run;
      value = tab->level;
      if (bs.GetBits(1))
        value = -value;
    }

#if 0  // TODO: reenable
#ifndef NDEBUG
  if (options.Trace_DCTCoeff)
    cout << "(" << run << "," << value << ") ";
#endif
#endif
  
  return false;
}


/************************************* initialization ********************************/

static void InitTAB(const DirectVLCTableEntry* src,DirectVLC* dest,int invalid_value)
{
  int maxbits=0;
  for (uint16 n=0; src[n].nBits>=0 ;n++)
    if (src[n].nBits>maxbits)
      maxbits=src[n].nBits;

  int size = (1<<maxbits);

  for (int i=0;i<size;i++)
    dest[i].value = invalid_value;

  for (uint16 n=0; src[n].nBits>=0 ;n++)
    if (src[n].nBits>0)
      {
        uint16 vlc_base = src[n].bits << (maxbits - src[n].nBits);
        
        for (int i=0;i<(1<<(maxbits - src[n].nBits));i++)
          {
            dest[vlc_base+i].value = n;
            dest[vlc_base+i].bits  = src[n].nBits;
          }
      }
}

static class DummyInit_23487635
{
public:
  DummyInit_23487635()
    {
      InitTAB(vlc_mbaddrinc_tab,vlc_mbaddr,-1);
      InitTAB(vlc_mbmode_P_tab,vlc_mbmode_P,0);
      InitTAB(vlc_mbmode_B_tab,vlc_mbmode_B,0);
      InitTAB(vlc_cbp_tab,vlc_cbp,-1);
      InitTAB(vlc_DClum_tab,vlc_DClum,-1);
      InitTAB(vlc_DCchrom_tab,vlc_DCchrom,-1);
      InitTAB(vlc_mcode_tab,vlc_mc,-1);
    }
} dummyinit_23845673;
